@@ -5,6 +5,9 @@ import tarfile
 import time
 import uuid
 from datetime import datetime
+from datetime import date
+from http.client import BadStatusLine
+from http.client import RemoteDisconnected
 from tempfile import gettempdir
 from tempfile import NamedTemporaryFile
 
@@ -16,7 +19,10 @@ from ibutsu_client import Configuration
 from ibutsu_client import HealthApi
 from ibutsu_client import ResultApi
 from ibutsu_client import RunApi
+from json import JSONEncoder
 from urllib3.exceptions import MaxRetryError
+from urllib3.exceptions import ProtocolError
+
 
 # A list of markers that can be filtered out
 FILTERED_MARKERS = ["parametrize"]
@@ -34,6 +40,17 @@ BLOCKER_CATEGORY_TO_CLASSIFICATION = {
 # Place a limit on the file-size we can upload for artifacts
 UPLOAD_LIMIT = 256 * 1024  # 256 KiB
 
+# Maximum number of times an API call is retried
+MAX_CALL_RETRIES = 3
+
+
+class DateTimeEncoder(JSONEncoder):
+    """ Handle datetime objects in the archiver. """
+
+    def default(self, obj):
+        if isinstance(obj, (date, datetime)):
+            return obj.isoformat()
+
 
 def safe_string(o):
     """This will make string out of ANYTHING without having to worry about the stupid Unicode errors
@@ -109,6 +126,10 @@ def overall_test_status(statuses):
     return "passed"
 
 
+class TooManyRetriesError(Exception):
+    pass
+
+
 class IbutsuArchiver(object):
     """
     Save all Ibutsu results to archive
@@ -122,7 +143,7 @@ class IbutsuArchiver(object):
         self._results = {}
         self._run_id = None
         self.run = None
-        self.temp_path = path
+        self._temp_path = path
         self.source = source or "local"
         self.extra_data = extra_data or {}
 
@@ -142,7 +163,7 @@ class IbutsuArchiver(object):
             )
         # If the project is set via environment variables
         if os.environ.get("IBUTSU_PROJECT"):
-            self.extra_data.update({"project", os.environ.get("IBUTSU_PROJECT")})
+            self.extra_data.update({"project": os.environ.get("IBUTSU_PROJECT")})
 
     def _status_to_summary(self, status):
         return {
@@ -154,15 +175,24 @@ class IbutsuArchiver(object):
             "tests": "tests",
         }.get(status)
 
+    def get_temp_path(self, run):
+        if not self._temp_path:
+            self._temp_path = os.path.join(gettempdir(), run["id"])
+            os.makedirs(self._temp_path, exist_ok=True)
+        return self._temp_path
+
+    @property
+    def temp_path(self):
+        if not self.run:
+            raise Exception("Run ID has not yet been set")
+        return self.get_temp_path(self.run)
+
     def _save_run(self, run):
-        if not self.temp_path:
-            self.temp_path = os.path.join(gettempdir(), run["id"])
-        os.makedirs(self.temp_path, exist_ok=True)
         if not run.get("metadata"):
             run["metadata"] = {}
         run["metadata"].update(self.extra_data)
-        with open(os.path.join(self.temp_path, "run.json"), "w") as f:
-            json.dump(run, f)
+        with open(os.path.join(self.get_temp_path(run), "run.json"), "w") as f:
+            json.dump(run, f, cls=DateTimeEncoder)
 
     @property
     def run_id(self):
@@ -258,7 +288,7 @@ class IbutsuArchiver(object):
             result["metadata"] = {}
         result["metadata"].update(self.extra_data)
         with open(os.path.join(art_path, "result.json"), "w") as f:
-            json.dump(result, f)
+            json.dump(result, f, cls=DateTimeEncoder)
         self._results[result_id] = result
         return result
 
@@ -269,7 +299,7 @@ class IbutsuArchiver(object):
             result["metadata"] = {}
         result["metadata"].update(self.extra_data)
         with open(os.path.join(art_path, "result.json"), "w") as f:
-            json.dump(result, f)
+            json.dump(result, f, cls=DateTimeEncoder)
         self._results[id] = result
 
     def upload_artifact(self, id, filename, data):
@@ -479,11 +509,17 @@ class IbutsuSender(IbutsuArchiver):
             if res.ready():
                 self._sender_cache.remove(res)
         try:
-            out = api_method(*args, **kwargs)
-            if "async_req" in kwargs:
-                self._sender_cache.append(out)
-            return out
-        except (MaxRetryError, ApiException) as e:
+            retries = 0
+            while retries < MAX_CALL_RETRIES:
+                try:
+                    out = api_method(*args, **kwargs)
+                    if "async_req" in kwargs:
+                        self._sender_cache.append(out)
+                    return out
+                except (RemoteDisconnected, ProtocolError, BadStatusLine):
+                    retries += 1
+            raise TooManyRetriesError('Too many retries while trying to call API')
+        except (MaxRetryError, ApiException, TooManyRetriesError) as e:
             self._has_server_error = self._has_server_error or True
             self._server_error_tbs.append(str(e))
             return None
